{- |Se define el tipo general de una expresion cualquiera, y se dan funciones que son utiles
 cualquiera que sea el area en que se usen. Los ejercicios tipo deben definirse directamente
 a partir de este modulo si no hay ningun otro del area considerada (ningun otro con los 
 operadores apropiados).
-}
module Expresion where
import List(elemIndices,nub)
import Maybe(fromJust)
import Monad(liftM)
import Algoritmos(vueltaAtras)
import Traza

{- |La expresion se concreta dando el tipo de valor v y el tipo de operador op.
 Por ejemplo, en logica podria ser v == Bool, op == AND|OR|XOR|Si|Sii.
 Este tipo se ampliara en el futuro admitiendo constantes, no solo variables y
 operaciones sobre ellas; por eso hay que especificar el tipo del valor, para evitar parte de
 las modificaciones que haya que hacer en el resto del programa debido a ese annadido.
 
 La implementacion no se oculta en este modulo, pero se dan 
 funciones pensando en una posible ocultacion desde otro modulo, que no exporte la 
 implentacion.

 Dos expresiones que solo se diferencien en el nombre de sus variables /son distintas/
-}
data Expresion v op = 
	V String | 			-- ^ Variable (con nombre)
	O op [Expresion v op] 		-- ^ Operacion
	deriving Eq


-- |Devuelve el operador mas externo de una expresion, o error si no hay operacion.
operador 	:: Expresion v op -> op
-- |Devuelve la lista de operandos, o error si no hay operacion.
operandos 	:: Expresion v op -> [Expresion v op]
-- |True sii la expresion es una variable
esVar 		:: Expresion v op -> Bool

{- |Compara una expresion y una referencia. Ejemplo: (x+y*z)*k tiene forma (a+b*c)*d, (a+b)*d, y (a*d); pero no tiene forma a+b.

Devuelve True sii la expresion tiene la forma dada en la referencia.
-}
tieneForma 	:: Eq op => 
	Expresion v op 		-- ^ expresion
	-> Expresion v op 	-- ^ referencia
	-> Bool


operador (O op _)	= op
operador (V _)		= error "Expresion.operador: no hay operación"

operandos (O _ l)	= l
operandos (V _)		= error "Expresion.operandos : no hay operación"

esVar (V _)			= True
esVar _				= False

e `tieneForma` (V _)		= True
(V _) `tieneForma` (O _ _)	= False
(O op x)`tieneForma`(O op' x')	
	|op /= op'		= False
	|otherwise  		= (and.map (\(a,b)-> a`tieneForma`b )) (zip x x') &&
		
			--que coincidan las repeticiones:
			if length x'== (length.nub) x' 
				--no hay repeticiones en patron:
				then True
				--si las hay; comprobar que en los mismos lugares:
				else lugarRepeticiones x == lugarRepeticiones x'		
	where
	lugarRepeticiones l = map (\e-> elemIndices e l) l

instance (Show v,Show op) => Show (Expresion v op) where
	show (V s)	= s
	show (O op l)	= show op ++ " " ++ (unwords.map show) l

----------------------
--Tranformaciones, para simplificar expresiones por axiomas
----------------------

-- | La forma deseada (la segunda) no debe tener ninguna variable que no aparezca en la primera.
type Transformacion v op = (Expresion v op, Expresion v op)

-- | Muestra la transformacion entre llaves {}. Analoga a una funcion show.
mostrarEq :: (Show v,Show op) => Transformacion v op -> String
mostrarEq (a,b) = "{("++ show a ++ ") -> (" ++ show b ++")}"

-- | Version mas completa de 'tieneForma'. En lugar de devolver tan solo un valor booleano,
-- devuelve la lista de transformaciones que habria que hacer en la referencia para que fuera
-- igual (no solo la misma forma) que la expresion dada.
tieneForma' :: Eq op 
	=> Expresion v op 		-- ^ expresion
	-> Expresion v op 		-- ^ referencia
	-> Maybe [Transformacion v op]	-- ^ transformaciones para igualar referencia a 
					-- expresion. 'Nothing' sii no tienen la misma forma.
e `tieneForma'` v@(V _)			= Just [(v,e)]
(V _) `tieneForma'` (O _ _)		= Nothing
(O op x)`tieneForma'`(O op' x')	
	|op /= op'			= Nothing
	|otherwise  			= let r=map (\(a,b)-> a`tieneForma'`b ) (zip x x') in
		if (and.map (/=Nothing)) r &&	-- todos tienen forma correcta
		 (length x'== (length.nub) x' || -- no hay repeticion en patron
		 lugarRepeticiones x == lugarRepeticiones x') --repes correctas en expresion
			then aux r
			else Nothing		
	where
	lugarRepeticiones l = map (\e-> elemIndices e l) l
	aux r = Just ((concat.map fromJust) r)

-- |Solamente se sustituyen variables. Mas general es 'simplificar'.
sustituir :: Eq op => 
	[Transformacion v op]	 	-- ^ sustituciones a realizar, forma [(variable,valor)].
					-- Las transformaciones de otra forma se ignoraran.
	-> Expresion v op 		-- ^ expresion a modificar
	-> Expresion v op		-- ^ salida: expresion modificada (si se dio valor para
					-- alguna de sus variables)
sustituir equivalencias e@(V p)		= fromJust( lookup e equivalencias )
sustituir equivalencias e@(O op l)	= O op (map (sustituir equivalencias) l)

-- |Simplifica con una sola transformacion dada. Es utilizada por 'simplificar'.
utilizarEquiv ::(Eq op,Show v,Show op) => 
	Transformacion v op 		-- ^ una unica transformacion (cualquier expresion a 
					-- cualquier expresion)
	-> Traza (Expresion v op) 	-- ^ una expresion dentro de una monada Traza 
	-> Traza (Expresion v op)	{- ^ salida: la misma expresion transformada (si se dio 
					 una transformacion para la forma de la expresion o 
					 de alguno de sus operandos). Ademas, traza:
					
					* la expresion anterior, 
					
					* la transformacion usada (axioma)
					
					* y el resultado.
					-}
utilizarEquiv axioma e
	|esVar (contenido e)		= e	
	|valoresForma2 /= Nothing	= do
				
		let axioma' = mostrarEq axioma
		let resultado = sustituir (fromJust valoresForma2) (snd axioma)
		trazar [show e ++ ", " ++ axioma' ++ " => " ++ show resultado]
		return resultado

	|otherwise 			= do

		e'<-e
		let l = map (utilizarEquiv axioma) (map return (operandos e'))
		let l' = map contenido l
		let resultado = O (operador (contenido e)) l'
		(trazar.concat) (map (memoria)  l)
		return resultado
	where
	valoresForma2 = (contenido e) `tieneForma'` (fst axioma)

-- | Simplifica una expresion atendiendo a una lista de tranformaciones posibles. La expresion
-- simplificada es devuelta en una monada traza que almacena las transformaciones hechas; esta 
-- memoria es legible por pantalla con solo aplicar la funcion 'traza' o 'trazaInversa'.
simplificar :: (Eq op,Show v,Show op) =>
	[Transformacion v op] -> Expresion v op -> [Traza (Expresion v op)]
simplificar axiomas e	= vueltaAtras valido hijos sinPodar (return e)
	where
	valido e = 	hijos e == [] --e no varia aplicando ningun axioma
	hijos e =	[e'|e'<-map (flip utilizarEquiv e) axiomas, e'/= e]
	sinPodar e = 	True
