-- | Implementa la monada Traza. Modela el efecto secundario de memorizar algunos datos.
module Traza(Traza,memoria,contenido,trazar,traza,trazaInversa) where

-- | Monada Traza, basada en la monada del articulo 'Introduction to functional programming' de
-- Philip Wadler <http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.47.4582>. Su lectura
-- es recomendable para entender los posibles usos de las monadas.
data Traza a 	= T {
		memoria::Memoria, 	-- ^ es el efecto secundario que estamos modelando
		contenido::a		-- ^ es el contenido explicito
		}
-- | Actualmente la memoria de la traza es una lista de String, (y no solo un String) para que -- sea facil ordenar de manera inversa.
type Memoria 	= [String]

-- | Ejecuta el efecto secundario
trazar		:: Memoria -> Traza()
-- | Devuelve la memoria como un solo String
traza 		:: Traza a -> String
-- | Devuelve la traza en orden contrario a 'traza'
trazaInversa	:: Traza a -> String


trazar t 	= T t ()
traza 		= unlines.memoria
trazaInversa	= unlines.reverse.memoria

-- | El comportamiento de la traza Monada es:
-- * >>=
-- ** contenido: solo se almacena el final
-- ** memoria: se concatenan (no se pierde nada)
-- * return. La memoria incial es una lista vacia. Recordar: si return esta al final de >>= (o bloque do), la memoria se ampliara concadenando el valor anterior a >>=

instance Monad Traza where
	m >>= k 	= let T x a = m in
			let T y b = k a in
			T (x++y) b
	
	return a 	= T [] a
	
-- | Solo se tiene en cuenta si el contenido es igual (no la memoria):
instance Eq a => Eq (Traza a) where
	a == b		= contenido a == contenido b

-- | Solo se muestra el contenido (no la memoria):
instance Show a => Show (Traza a) where
	show 		= show.contenido 
